from Acquisition import aq_parent

from plonehrm.absence import AbsenceMessageFactory as _
from plonehrm.absence.utils import localize

from DateTime import DateTime


def save_note(object, event):
    """ This event is fired when an absence note is created
    or updated. If a nextContactDate has been specified in the
    note, then a new item is created in the checklist.
    """

    # If no date for the next contact has been set, then the event
    # does not add anything in the checklist.
    if not object.next_contact_date:
        return

    # The employee is the grandfather of the note (father is absence).
    employee = aq_parent(aq_parent(object))

    # We get the employee checklist.
    content_filter = {'portal_type': 'Checklist'}
    contents = list(employee.getFolderContents(contentFilter=content_filter))

    # This shall not append as every employee has his own checklist.
    if len(contents) == 0:
        return

    checklist = contents[0].getObject()

    # We will be looking for a possibly translated string.
    title = _('msg_next_call_back_on', default=u'Call to query about absence')
    title = localize(object, title)

    # We create a link to add a new note in the current absence.

    # First we get the current absence.
    content_filter = {'portal_type': 'Absence'}
    absences = list(employee.getFolderContents(contentFilter=content_filter))

    # Normally, the current absence is the first in the list, so
    # the loop should be stopped at first iteration.
    currentAbsence = None
    for ab in absences:
        ab = ab.getObject()
        if ab.is_current_absence:
            currentAbsence = ab
            break

    link_url = currentAbsence.absolute_url() + '/createObject?type_name=Note'
    link_title = _('msg_new_note_in_absence',
                   default=u'Create a new note in the current absence')
    link_title = localize(object, link_title)

    # We cast python datetime to zope DateTime.
    d1 = object.next_contact_date
    date = DateTime(d1.year, d1.month, d1.day, d1.hour, d1.minute)

    # First, we check if there exists a previous item in the checklist
    # which tells to call the employee back.
    items = checklist.findItems(title, startswith = True)

    # If such an item exists, we delete it.
    for item in items:
        checklist.deleteItem(item)

    # We add the new item to the checklist.
    checklist.addItem(title,
                      date=date,
                      link_url = link_url,
                      link_title = link_title)



from datetime import date
from kss.core import kssaction
from plone.app.kss.plonekssview import PloneKSSView
from Acquisition import Explicit
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import getSiteEncoding
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.interface import Interface

from plonehrm.absence.interfaces import IAbsenceAdapter
from plonehrm.absence import AbsenceMessageFactory as _
from plonehrm.absence.absence import AbsenceDateError

MAXITEMS = 3


class IAbsenceListViewlet(Interface):

    def get_absence_ist():
        """Return a list of (maximum is MAXITEMS) absences
        """

    def show_absence_form():
        """Return true if the absence form should be shown
        """

    def show_absence_list():
        """Return true if the absence list should be shown
        """

    def get_current_absence():
        """Return the current absence, if any.
        """


class IAddAbsenceView(Interface):
    """Marker interface to add absencelist
    """


def simple_add_absence(context, text, start_date=None):
    if text is None or text =='':
        message = _(u'msg_no_input', default=u'No input provided')
    else:
        absencelist = IAbsenceAdapter(context)
        if not isinstance(text, unicode):
            text = unicode(text, getSiteEncoding(context))
        if start_date is None:
            start_date = date.today()
        try:
            absencelist.add_absence(text, start_date)
            message = None
        except AbsenceDateError, exc:
            message = exc
    return message


def simple_close_absence(context, length=None, end_date = None):
    """ close the open absence
    """
    absencelist = IAbsenceAdapter(context)
    if end_date is None:
        end_date = date.today()
    try:
        absencelist.close_absence(end_date, length)
        message = None
    except AbsenceDateError, exc:
        message = exc
    return message


class AddAbsence(BrowserView):
    """Add an Absence.

    Note that this is the non-kss version.  We may want to concentrate
    on the kss version instead (see below).
    """

    def __call__(self):
        """Create a new Absence if the inline form is used.

        Only react if the form was submitted
        Only react if someone filled in text for the Absence
        Make a new Absence with the correct text and date.
        When not set, the date will be set to today.
        """
        text = self.request.get('text', None)
        start_date = self.request.get('new_absence_date', None)
        if start_date:
            year = int(self.request.get('new_absence_date_year'))
            month = int(self.request.get('new_absence_date_month'))
            day = int(self.request.get('new_absence_date_day'))
            start_date = date(year, month, day)
        else:
            start_date = date.today()
        message = simple_add_absence(self.context, text, start_date)
        self.context.plone_utils.addPortalMessage(message)
        response = self.request.response
        here_url = self.context.absolute_url()
        response.redirect(here_url)


class AbsenceViewlet(Explicit):
    """A viewlet for seeing and adding absencelist.
    """

    template = ViewPageTemplateFile('absence_viewlet.pt')

    def __init__(self, context, request, view=None, manager=None):
        self.context = context
        self.request = request
        self.view = view
        self.form_state = False
        self.close_form_state = False
        self.error_message = None

        self.manager = manager
        membership = getToolByName(self.context, 'portal_membership')
        self.checkPermission = membership.checkPermission
        self.encoding = getSiteEncoding(context)

    def get_absence_list(self):
        absences = IAbsenceAdapter(self.context)
        absencelist = absences.absencelist
        result = absencelist[-MAXITEMS:]
        result.reverse()
        return result

    def get_current_absence(self):
        absences = IAbsenceAdapter(self.context)
        return absences.current_absence()

    def show_absence_list(self):
        if self.form_state or self.close_form_state:
            return False
        return True

    def allow_adding_absence(self):
        if not self.checkPermission('plonehrm: Add absence', self.context):
            return False
        if IAbsenceAdapter(self.context).is_absent():
            return False
        return True

    def show_close_absence_button(self):
        return IAbsenceAdapter(self.context).is_absent()

    def show_add_absence_button(self):
        if not self.allow_adding_absence():
            return False
        return not self.form_state

    def show_absence_form(self):
        if not self.allow_adding_absence():
            return False
        return self.form_state

    def show_absence_close_form(self):
        return self.close_form_state

    def render(self):
        val = self.template()
        #if not isinstance(val, unicode):
        #    val = unicode(val, self.encoding)
        return val


class AbsenceList(BrowserView):
    """A view for listing the complete absencelist of an employee."""

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_absence_list(self):
        absences = IAbsenceAdapter(self.context)
        absencelist = absences.absencelist
        # Create a copy and reverse that copy, instead of mistakenly
        # reversing the original.
        result = absencelist[:]
        result.reverse()
        return result


class AbsenceAddView(PloneKSSView):
    """kss view for adding a absence
    """

    @kssaction
    def add_absence(self, text, start_date, start_date_year,
                    start_date_month, start_date_day):
        """Add a absence
        """
        core = self.getCommandSet('core')
        selector = core.getHtmlIdSelector('absencelist')
        if start_date:
            year = int(start_date_year)
            month = int(start_date_month)
            day = int(start_date_day)
            start_date = date(year, month, day)
        else:
            start_date = date.today()

        message = simple_add_absence(self.context, text, start_date)
        view = self.context.restrictedTraverse('@@plonehrm.absence')

        if message:
            view.error_message = message
            view.form_state = True
        rendered = view.render()
        core.replaceHTML(selector, rendered)


class AbsenceCloseView(PloneKSSView):
    """kss view for closing a absence
    """

    @kssaction
    def close_absence(self, length, end_date, end_date_year,
                      end_date_month, end_date_day):
        """Close the open absence"""
        core = self.getCommandSet('core')
        selector = core.getHtmlIdSelector('absencelist')
        if end_date:
            year = int(end_date_year)
            month = int(end_date_month)
            day = int(end_date_day)
            end_date = date(year, month, day)
        else:
            end_date = date.today()
        try:
            absence_length = int(length)
        except ValueError:
            message = _(u'msg_days_should_be_integer',
                        default=u'The number of days should be an integer')
        else:
            message = simple_close_absence(self.context, absence_length,
                                           end_date)
        view = self.context.restrictedTraverse('@@plonehrm.absence')
        if message:
            view.error_message = message
            view.close_form_state = True
        rendered = view.render()
        core.replaceHTML(selector, rendered)


class AbsenceKssView(PloneKSSView):
    """kss view for managing the absence viewlet"""

    @kssaction
    def kss_absence_control(self, dostate):

        core = self.getCommandSet('core')
        selector = core.getHtmlIdSelector('absencelist')

        view = self.context.restrictedTraverse('@@plonehrm.absence')

        if dostate == "absenceform":
            view.form_state = True
        else:
            view.form_state = False


        if dostate == "closeabsenceform":
            view.close_form_state = True
        else:
            view.close_form_state = False

        rendered = view.render()
        core.replaceHTML(selector, rendered)
        # Set the focus on the input field, which also clears the
        # previous text entered.
        if view.form_state == True:
            core.focus('#new-absence-text')



"""Definition of the Absence content type
"""

from datetime import date
from Acquisition import aq_chain
from zope.interface import implements
from Products.Archetypes import atapi
from Products.ATContentTypes.content import folder
from Products.ATContentTypes.content import schemata

from plonehrm.absence import AbsenceMessageFactory as _
from plonehrm.absence.interfaces import IAbsence
from Products.plonehrm.interfaces import IEmployee
from plonehrm.absence.config import PROJECTNAME
from zope import component

AbsenceSchema = folder.ATFolderSchema.copy() + atapi.Schema((

    atapi.DateTimeField(
        name='startDate',
        widget=atapi.CalendarWidget(
            label=_(u'label_start_date', default=u'Start date'),
            show_hm = False,
        )
    ),

    atapi.DateTimeField(
        name='endDate',
        widget=atapi.CalendarWidget(
            label=_(u'label_end_date', default=u'End date'),
            show_hm = False,
        )
    ),

    atapi.IntegerField(
        name='absence_length',
        widget=atapi.IntegerWidget(
            label=_(u'label_days_absent', default=u'Absence length (days)'),
        )
    ),

))

# Set storage on fields copied from ATFolderSchema, making sure
# they work well with the python bridge properties.

AbsenceSchema['title'].storage = atapi.AnnotationStorage()
AbsenceSchema['description'].storage = atapi.AnnotationStorage()
AbsenceSchema['description'].widget.visible = {'view': 'visible',
                                               'edit': 'invisible'}
schemata.finalizeATCTSchema(
    AbsenceSchema,
    folderish=True,
    moveDiscussion=False,
    )


class Absence(folder.ATFolder):
    """Absence"""
    implements(IAbsence)

    meta_type = "Absence"
    schema = AbsenceSchema

    title = atapi.ATFieldProperty('title')
    description = atapi.ATFieldProperty('description')
    # XXX Make 'text' a rich text field?
    text = atapi.ATFieldProperty('title')
    start_date = atapi.ATDateTimeFieldProperty('startDate')
    end_date = atapi.ATDateTimeFieldProperty('endDate')

    def __repr__(self):
        value = "<Absence: %s" % self.text
        if self.start_date:
            # We get the (python) date instead of the datetime here.
            value += ", " + self.start_date.date().isoformat()
        if self.end_date:
            value += " - " + self.end_date.date().isoformat()
        return value + ">"

    def get_employee(self):
        """Get the employee that this contract is in.

        Note that we probably are in the portal_factory, so our direct
        parent is not an Employee.  But we can traverse up the
        acquisition chain to find it.
        """
        for parent in aq_chain(self):
            if IEmployee.providedBy(parent):
                return parent

    def days_absent(self):
        """How long did the absence take?

        If the absence has ended, return the length stored on the
        object. Else try to estimate the number of days someone hasn't
        worked.

        If the employee works a 0 hour/day contract the employee will get
        0 sick days by his/her name when they are sick or were sick.
        """
        if self.end_date:
            return self.get('absence_length', 0)

        parent = self.get_employee()
        if parent is None:
            return
        view = component.getMultiAdapter((parent, self.REQUEST),
                                         name=u'contracts')
        current = view.current_contract()
        if current is not None:
            days_per_week = current.getDaysPerWeek()
        else:
            days_per_week = 5

        end_date = self.end_date
        if not end_date:
            end_date = date.today()

        absence_length = float(end_date.toordinal() -
                               self.start_date.toordinal() + 1)
        sickness_days = (days_per_week / 7.0) * absence_length
        rounded_sickness = int(round(sickness_days))
        return rounded_sickness

    @property
    def is_current_absence(self):
        """ Returns True is the absence is not closed and
        False in the other cases.
        """
        return not bool(self.end_date)

atapi.registerType(Absence, PROJECTNAME)



from Acquisition import Explicit
from zope.component import adapter

from plonehrm.absence.notifications.interfaces import IAbsenceEvent


class MockAQ(Explicit):

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class MockMailer(MockAQ):

    def __init__(self, employee, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.employee = employee

    def send(self):
        print self.template.pt_source_file(), self.employee, self.subject


@adapter(IAbsenceEvent)
def event_tracker(event):
    print event



from zope.interface import Interface


class IAbsence(Interface):
    """Absence"""


class IAbsenceFile(Interface):
    """Absence File"""


class IAbsenceAdapter(Interface):
    """Interface for the Absence adapter.
    This is used to get/set absence information on Employees.
    """

    def add_absence(text, start_date=None):
        """Add a new absence to the absencelist."""

    def close_absence(end_date):
        """Close the currently open absence."""

    def remove_absence(index):
        """Remove absence ``index`` from the absencelist."""

    def current_absence():
        """Return the unterminated absence."""

    def is_absent():
        """Test for an unterminated absence."""


class INote(Interface):
    """Notes to add to an absence"""



from datetime import date

from interfaces import IAbsenceAdapter
from zope.interface import implements
from zope.event import notify
from zope import component

from Products.Archetypes.event import ObjectInitializedEvent
from Products.plonehrm.interfaces import IEmployee
from plonehrm.absence.flags import NOTIFICATION_PREFIX
from plonehrm.notifications.interfaces import INotified
from plonehrm.absence import AbsenceMessageFactory as _


class AbsenceDateError(Exception):

    def __str__(self):
        return u"Date Error"


class ConflictingStartDateError(AbsenceDateError):

    def __str__(self):
        return _('msg_conflicting_start_date', u"Error: Conflicting start date")


class StartDateInFutureError(AbsenceDateError):

    def __str__(self):
        return _('msg_start_date_in_future', u"Error: Start date in future")


class CloseDateBeforeStartDateError(AbsenceDateError):

    def __str__(self):
        return _('msg_close_date_before_start_date', u"Error: Close date before start date")


class AbsenceAdapter(object):
    """This adapter can be used to get/set absence information on Employees.
    """
    implements(IAbsenceAdapter)
    component.adapts(IEmployee)

    def __init__(self, context):
        self.context = context

    @property
    def absencelist(self):
        contentFilter = {'portal_type': 'Absence'}
        brains = self.context.getFolderContents(contentFilter=contentFilter)
        objects = [brain.getObject() for brain in brains]

        def date_key(item):
            return item['start_date']

        items = sorted(objects, key=date_key)
        return items

    def add_absence(self, text, start_date=None):
        """Add a new absence to the absencelist."""
        assert(not self.is_absent()), \
            'there can only be one open absence'
        assert(isinstance(text, basestring)), \
            'string expected, got %s' % type(text)
        assert(start_date is None or
               isinstance(start_date, date)), \
            'date expected, got %s' % type(start_date)

        # Make sure we have a valid start date.
        if start_date is None:
            start_date = date.today()
        if (self.absencelist and
            start_date.toordinal() <= \
                self.absencelist[-1].end_date.toordinal()):
            raise ConflictingStartDateError
        # No start date in future please.  Well... tomorrow is okay.
        if start_date.toordinal() > date.today().toordinal() + 1:
            raise StartDateInFutureError

        #absence = Absence(text, start_date)
        #self.absencelist.append(absence)
        #self.absencelist.sort()

        new_id = self.context.generateUniqueId('Absence')
        self.context.invokeFactory('Absence',
                                   id=new_id,
                                   title=text)
        new_absence = getattr(self.context, new_id)
        new_absence.start_date = start_date
        new_absence.unmarkCreationFlag()
        new_absence._renameAfterCreation()
        notify(ObjectInitializedEvent(new_absence))

        return new_absence

    def close_absence(self, end_date, length=None, text=None):
        """Close the currently open absence."""
        assert(isinstance(end_date, date)), \
            'date expected, got %s' % type(end_date)
        absence = self.current_absence()
        assert(absence)
        if end_date.toordinal() < absence.start_date.toordinal():
            raise CloseDateBeforeStartDateError
        absence.end_date = end_date
        if text:
            absence.text = text
        if length is not None:
            absence.absence_length = length
        else:
            absence.absence_length = absence.days_absent()
        notified = INotified(self.context)
        notified.remove_notification_set(NOTIFICATION_PREFIX)
        return absence

    def remove_absence(self, index):
        """Remove absence ``index`` from the absencelist."""
        assert(len(self.absencelist) >= index + 1)
        to_remove = self.absencelist.pop(index)
        self.context._delObject(to_remove.id)
        return to_remove

    def current_absence(self):
        """Return the unterminated absence."""
        for absence in self.absencelist:
            if not absence.end_date:
                return absence

    def is_absent(self):
        """Test for an unterminated absence."""
        if self.current_absence():
            return True
        return False
